#include <iostream>   // cout
#include <iomanip> 
#include <vector> 
#include <fstream> 
#include <stdexcept>
#include <stdlib.h> 

#include "strategy.h" 
#include "my_strategy.h" 


using namespace std; 

#define DRAND()	  (((double) rand()) / ( RAND_MAX + 1.0))


#define TRIALS 100
#define PRINT_TRIALS 30  

int rewards[2][2] = { { CC, CD }, { DC, DD } }; 

#define N 6 // number of strategies 
string stratNames[N] = {
  "RandomStrategy", 
  "AlwaysCooperate", 
  "AlwaysDeceive",
  "TFT (CopyPartner)", 
  "Alternate", 
  "Pavlov"}; 


Strategy* createStrategy(const string& name){
  
  if (name == "RandomStrategy") return new RandomStrategy(); 
  if (name == "AlwaysCooperate") return new AlwaysCooperate(); 
  if (name == "AlwaysDeceive") return new AlwaysDeceive();
  if (name == "TFT (CopyPartner)") return new CopyPartner();
  if (name == "Alternate") return new Alternate();
  if (name == "Pavlov") return new Pavlov();

  throw std::runtime_error("unknown strategy: "+name);
  return 0;
} 

class result {
public:
  int first_player;
  int second_player; 
};  

result compete(Strategy& strat1, Strategy& strat2) {

  int res1, res2;
  int rec1[TRIALS];
  int rec2[TRIALS]; 

  cout << strat1.getName() << " vs. " << strat2.getName() << endl; 

  strat1.reset(); strat2.reset(); // prepare for competition 

  result res; 
  res.first_player = 0;
  res.second_player = 0; 

  // compete 
  for (int i=0; i<TRIALS; i++) {
   
    // ask for decision 
    rec1[i] = res1 = strat1(); 
    rec2[i] = res2 = strat2();
    

    // some strategies returns illegal value 
    if ((res1 != COOPERATE) && (res1 != DECEIVE)) {
      if (DRAND()>0.5)
	rec1[i] = res1 = COOPERATE;
      else
	rec1[i] = res1 = DECEIVE;
    } 

    if ((res2 != COOPERATE) && (res2 != DECEIVE)) {
      if (DRAND()>0.5)
	rec2[i] = res2 = COOPERATE;
      else
	rec2[i] = res2 = DECEIVE;
    } 


    // rewards
    strat1.reward( rewards[res1][res2] ); 
    strat2.reward( rewards[res2][res1] ); 
  
    res.first_player += rewards[res1][res2]; 
    res.second_player += rewards[res2][res1]; 
  }

  cout << strat1.getName() << " vs. " << strat2.getName() << endl; 
  cout <<"Player1:  ";
  for (int i=0; i<PRINT_TRIALS; i++) cout << ((rec1[i]) ? "D" : "C");
  cout << " => " << res.first_player << endl; 
  cout << "Player2:  "; 
  for (int i=0; i<PRINT_TRIALS; i++) cout << ((rec2[i]) ? "D" : "C"); 
  cout << " => " << res.second_player << endl << endl; 

  
  return res; 
} 


int main() {

  srand( time(0) );

  // array of strategies 
  vector<Strategy*> strats; 
  strats.resize(N);

  for (int i=0; i<N; i++)
    strats[i] = createStrategy( stratNames[i] ); 

  int results[N][N]; 
  int sum_results[N]; 
  for (int i=0; i<N; i++) sum_results[i] = 0; 

  result res; 

  // compete 
  for (int i=0; i< N; i++)  
    for (int j=0; j<i; j++) {
      res = compete( *strats[i], *strats[j] ); 
      
      results[i][j] = res.first_player;
      results[j][i] = res.second_player;
     
      sum_results[i] += results[i][j]; 
      sum_results[j] += results[j][i]; 
    } 
  
  // compete with myself
  for (int i=0; i<N; i++) {
    Strategy* myCopy = createStrategy( strats[i]->getName() );
    
     res = compete( *strats[i], *myCopy); 
      
     results[i][i] = (res.first_player + res.second_player)/2;
     
     sum_results[i] += results[i][i]; 

     delete myCopy;
  } 

  /*
  // print results
  for (int i=0; i<N; i++) {
    for (int j=0; j<N; j++) {
      cout <<  results[i][j] << " "; 
    } 
    cout << endl; 
  } 
  */

  for (int i=0; i<N; i++) 
    cout << setw(30) << strats[i]->getName() << ":  " << sum_results[i]/N << endl; 
  
  ofstream fout("results.txt");
  for (int i=0; i<N; i++) 
    fout << sum_results[i]/N << " " << strats[i]->getName() << endl; 
  
  fout.close();

  for (unsigned i=0; i<strats.size(); i++) 
    delete strats[i]; 

  return 0; 
} 
